from collections import deque
from .q_function import QFunction


class Memory:
    def __init__(self, size):
        self._input_records = deque(maxlen=size)
        self._signal_records = deque(maxlen=size)

    @property
    def is_full(self):
        return len(self._input_records) == self._input_records.maxlen

    def record(self, ann_input, signal):
        self._input_records.append(ann_input)
        self._signal_records.append([signal])

    @property
    def inputs(self):
        return self._input_records

    @property
    def signals(self):
        return self._signal_records


class QNeuronal(QFunction):
    def __init__(self, ann, n, memory_size=None):
        """
        This class implements a nonlinear Q function approximation using an artificial neural network. The ANN is
        trained on the environment state and the Q-Value signals to be able to predict Q-Values given a specific state.

        States are provided to the ANN as a list of lists, i.e.:
        ``[[0, 0], [0, 1], [0, 1]]``
        Signals are provided to the ANN as list of of single element lists, i.e.:
        ``[[0.5], [-0.9], [1.0]]``

        :param ann: Artificial Neural Network to be used for Q-Value prediction.
        :param n: Number of actions
        :param memory_size: (optional) Size of recorded state, signal pairs that are passed to the ANN to as a batch.
        """
        super().__init__(list(range(n)))
        self.ann = ann
        self.memory = Memory(1 if memory_size is None else memory_size)

    def __getitem__(self, state_action):
        return self.ann.predict([self._make_input_state(state_action)])[0][0]

    def learn(self, state, action, signal):
        self.memory.record(self._make_input_state((state, action)), signal)
        if self.memory.is_full:
            self.ann.train(self.memory.inputs, self.memory.signals)



from .neural_network import NeuralNetwork
from .q_model import QModel


class QRegressionModel(QModel):
    def __init__(self, input_size, hidden_layers, lr=0.01, seed=None):
        """
        Implements a regression Artificial Neural Network using linear rectifier units as hidden neurons and outputs a
        linear float value. Training is done using a gradient descent and the loss is calculated using the mean squared
        error. Training and target batches are expected as a list of lists.

        :param input_size: Size of the input state vector
        :param hidden_layers: List specifying number and size of hidden relu neuron layers, i.e.: ``[3, 2]`` specify two hidden layers of size 3 and 2
        :param lr: Learning rate parameter
        :param seed: Seed used by random operations
        """
        QModel.__init__(self, input_size)
        self.ann = NeuralNetwork(input_size, seed)
        for units in hidden_layers:
            self.ann.add_layer(units, activation='relu', weight_init='glorot_uniform')
        self.ann.add_layer(1, activation='linear', weight_init='glorot_uniform')
        self.ann.compile('mean_squared_error', 'gradient_descent', lr)

    def do_prediction(self, state):
        return self.ann.predict(state)

    def do_training(self, states, targets):
        self.ann.train(states, targets)



import tensorflow as tf

ACTIVATIONS = {
    'linear': tf.identity,
    'relu': tf.nn.relu
}


def _activation_func(activation):
    func = ACTIVATIONS.get(activation)
    if func is None:
        raise NotImplementedError("The specified activation function '{}' has not been implemented".format(activation))
    return func


LOSSES = {
    'mean_squared_error': tf.losses.mean_squared_error
}


def _loss_func(loss):
    func = LOSSES.get(loss)
    if func is None:
        raise NotImplementedError("The specified loss function '{}' has not been implemented".format(loss))
    return func


OPTIMIZERS = {
    'gradient_descent': tf.train.GradientDescentOptimizer
}


def _optimizer(optimizer):
    cls = OPTIMIZERS.get(optimizer)
    if cls is None:
        raise NotImplementedError("The specified optimizer '{}' has not been implemented".format(optimizer))
    return cls


class NeuralNetwork:
    def __init__(self, input_size, seed=None):
        """
        Implements a sequenced neural network using Tensorflow. With ``add_layer`` you can add hidden neuron layers with
        specific activation functions and initializations. With ``compile`` you finish your setup and specify the loss
        and optimization function. The ``train`` function lets you fit your neural network to training data, while
        ``predict`` uses the current state of the neural network to predict values of given states.

        :param input_size: size of the input vector
        :param seed: (optional) a seed used for initialization
        """
        self._inputs = tf.placeholder(tf.float32, [None, input_size])
        self._last_layer = self._inputs
        self._last_units = input_size
        self._layer_idx = 0
        self._targets = None
        self._train_step = None

        self._initializers = {
            'ones': tf.ones_initializer(),
            'zeros': tf.zeros_initializer(),
            'glorot_uniform': tf.glorot_uniform_initializer(seed)
        }

    def add_layer(self, units, activation, weight_init, bias_init='zeros'):
        if self._train_step is not None:
            raise InvalidOperationError("Adding layers after compiling a network is not supported")
        W = self._make_weights(weight_init, units)
        b = self._make_bias(bias_init, units)
        self._last_layer = _activation_func(activation)(tf.matmul(self._last_layer, W) + b)
        self._last_units = units
        self._layer_idx += 1

    def _make_weights(self, init, units):
        return tf.get_variable("W{}".format(self._layer_idx), shape=[self._last_units, units],
                               initializer=self._initializer(init))

    def _initializer(self, init):
        func = self._initializers.get(init)
        if func is None:
            raise NotImplementedError("The specified initialization '{}' has not been implemented".format(init))
        return func

    def _make_bias(self, bias_init, units):
        return tf.get_variable("b{}".format(self._layer_idx), shape=[units], initializer=self._initializer(bias_init))

    def compile(self, loss, optimizer, learning_rate):
        if self._layer_idx == 0:
            raise InvalidOperationError("A Neural Network needs at least one layer to be compiled")
        self._targets = tf.placeholder(tf.float32, shape=[None, self._last_units])
        loss_func = _loss_func(loss)(self._targets, self._last_layer)
        self._train_step = _optimizer(optimizer)(learning_rate).minimize(loss_func)

        tf.global_variables_initializer().run()

    def predict(self, inputs):
        return self._last_layer.eval(feed_dict={self._inputs: inputs})

    def train(self, inputs, targets):
        if self._train_step is None:
            raise NotCompiledError("The network needs to be compiled before it can be trained")
        self._train_step.run(feed_dict={self._inputs: inputs, self._targets: targets})


class NotCompiledError(Exception):
    pass


class InvalidOperationError(Exception):
    pass


import random

import numpy as np


class Policy:
    def __init__(self, action_filter):
        self.action_filter = action_filter

    def _get_actions(self, state, q_function):
        if self.action_filter is None:
            return q_function.action_space

        return self._filter_invalid_actions(state, q_function)

    def _filter_invalid_actions(self, state, q_function):
        def remove_state(state_action):
            s, a = state_action
            return a

        return list(
            map(remove_state, filter(self.action_filter, map(lambda a: (state, a), q_function.action_space))))

    @staticmethod
    def _get_q_values_of_state(state, actions, q_function):
        return [q_function[state, a] for a in actions]


class EpsilonGreedyPolicy(Policy):
    def __init__(self, epsilon, action_filter=None):
        super().__init__(action_filter)
        self.epsilon = epsilon

    def select(self, state, q_function):
        if callable(self.epsilon):
            e = self.epsilon()
        else:
            e = self.epsilon

        action_space = self._get_actions(state, q_function)
        if random.random() < e:
            return random.choice(action_space)

        vs = self._get_q_values_of_state(state, action_space, q_function)
        return action_space[np.argmax(vs)]


class NormalEpsilonGreedyPolicy(Policy):
    def __init__(self, epsilon, action_filter=None):
        super().__init__(action_filter)
        self.epsilon = epsilon

    def select(self, state, q_function):
        if callable(self.epsilon):
            e = self.epsilon()
        else:
            e = self.epsilon

        action_space = self._get_actions(state, q_function)
        vs = self._get_q_values_of_state(state, action_space, q_function)
        return action_space[np.argmax(vs + np.random.randn(1, len(action_space)) * e)]





class QFunction:
    def __init__(self, action_space):
        self.action_space = action_space

    def _make_input_state(self, state_action):
        state, action = self._retrieve_state_action(state_action)
        state = self.box_state(state)

        state.append(action)
        return state

    def _retrieve_state_action(self, state_action):
        state, action = state_action
        self.check_action(action)
        return state, action

    @staticmethod
    def box_state(state):
        if not isinstance(state, list):
            state = [state]
        return list(state)

    def check_action(self, action):
        if action not in self.action_space:
            raise InvalidAction("The action {} is not part of action space {}".format(action, str(self.action_space)))


class InvalidAction(Exception):
    pass



class Formats:
    REWARD = '\033[94m'
    HIGHLIGHT = '\033[91m'
    END = '\033[0m'


class ValueTable(object):
    def __init__(self, initializer=None):
        if initializer is None:
            initializer = self.zero_init
        self.initializer = initializer
        self.storage = {}

    @staticmethod
    def zero_init():
        return 0

    def get_state(self, state):
        key = self.to_key(state)
        if key not in self.storage:
            return self.initializer()
        return self.storage[key]

    @staticmethod
    def to_key(state):
        return ' '.join(map(str, state))

    def update(self, state, value):
        key = self.to_key(state)
        self.storage[key] = value

    def print_states(self, state_sequences):
        print_strings = []
        for states in state_sequences:
            states_strings = []
            max_value = -float("inf")
            max_idx = 0
            for i, state in enumerate(states):
                key = self.to_key(state)
                if key in self.storage:
                    value = self.storage[key]
                    states_strings.append(
                        "{" + str(state) + ", " + Formats.REWARD + "{:6.2f}".format(
                            value) + Formats.END + "}")
                    if value > max_value:
                        max_value = value
                        max_idx = i
                else:
                    print("The state", state, "hasn't been set yet.")
                    return

            states_strings[max_idx] = states_strings[max_idx].replace(Formats.REWARD, Formats.HIGHLIGHT)
            print_strings.append(', '.join(states_strings))

        print('\n'.join(print_strings))

    def __getitem__(self, state):
        return self.get_state(state)

    def __setitem__(self, state, value):
        self.update(state, value)

    def print_all(self):
        return self.print_all_sorted_by(None)

    def print_all_sorted_by(self, state_index):
        out_strings = []
        sorter = None
        if state_index is not None:
            sorter = lambda t: t.split()[state_index]

        for key in sorted(self.storage, key=sorter):
            out_strings.append("[" + str(key) + "] = " + str(self.storage[key]))

        print('\n'.join(out_strings))



from .q_function import QFunction
from .value_table import ValueTable


class QTable(QFunction):
    def __init__(self, action_space, initializer=None):
        super().__init__(action_space)
        self.storage = ValueTable(initializer)

    def __getitem__(self, state_action):
        state = self._make_input_state(state_action)
        return self.storage[state]

    def __setitem__(self, state_action, value):
        state = self._make_input_state(state_action)
        self.storage[state] = value

    def learn(self, state, action, signal):
        self[state, action] += signal



from abc import ABCMeta, abstractmethod


class QModel(object, metaclass=ABCMeta):
    def __init__(self, input_size):
        self.input_size = input_size

    def predict(self, state):
        self.check_state(state[0])
        return self.do_prediction(state)

    @abstractmethod
    def do_prediction(self, state):
        pass

    def check_state(self, state):
        state_size = len(state)
        if state_size != self.input_size:
            raise InvalidDataSize(
                "Input size required is {}, but received state has length {}".format(self.input_size, state_size))

    def train(self, states, targets):
        self.check_state(states[0])
        self.check_target(targets[0])
        self.check_input_arrays(states, targets)

        self.do_training(states, targets)

    @abstractmethod
    def do_training(self, states, targets):
        pass

    @staticmethod
    def check_target(target):
        target_size = len(target)
        if target_size != 1:
            raise InvalidDataSize(
                "Output size of the model is {}, the training model_data has length {}".format(1, target_size))

    @staticmethod
    def check_input_arrays(states, targets):
        num_states = len(states)
        num_targets = len(targets)
        if num_states != num_targets:
            raise InvalidDataSize(
                "The number of states ({}) differs from the numbers of targets ({})".format(num_states, num_targets))


class InvalidDataSize(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)



from collections import deque
from functools import reduce


class Records:
    def __init__(self, size):
        self.size = size
        self._state_actions = deque()
        self._rewards = deque()

    def record_state_action(self, s, a):
        self._state_actions.append((s, a))

    def record_reward(self, r):
        self._rewards.append(r)

    @property
    def rewards(self):
        return self._rewards

    @property
    def is_full(self):
        return len(self._state_actions) == self.size

    @property
    def is_empty(self):
        return len(self._state_actions) == 0

    def __iter__(self):
        return self

    def __next__(self):
        self._rewards.popleft()
        return self._state_actions.popleft()

    def clear(self):
        self._state_actions.clear()
        self._rewards.clear()


class TDAgent:
    def __init__(self, policy, q_function, n, gamma, alpha):
        """
        Temporal difference learning agent, that implements an n-step sarsa algorithm

        :param policy: action selection policy implementation to be used
        :param q_function: the Q value function used to learn and predict the Q value
        :param n: number of steps the TD algorithm takes
        :param gamma: discounting factor of future rewards
        :param alpha: step size of signal
        """
        self.policy = policy
        self.q = q_function
        self.n = n
        self.gamma = gamma
        self.alpha = alpha

        self._record = Records(self.n)

    def start(self, state):
        self._record.clear()
        a = self.policy.select(state, self.q)
        return self._take_and_record(state, a)

    def _take_and_record(self, s, a):
        self._record.record_state_action(s, a)
        return a

    def step(self, state, reward):
        self._record.record_reward(reward)
        action = self.policy.select(state, self.q)
        if self._record.is_full:
            self._learn(state, action)

        return self._take_and_record(state, action)

    def _learn(self, s_next, a_next):
        g = self._calc_recorded_rewards_sum()
        g += pow(self.gamma, self.n) * self.q[s_next, a_next]
        s, a = next(self._record)
        self.q.learn(s, a, self._calc_signal(s, a, g))

    def _calc_recorded_rewards_sum(self):
        def decaying_sum(acc, enum):
            i, r = enum
            return acc + pow(self.gamma, max(i, 0)) * r

        return reduce(decaying_sum, enumerate(self._record.rewards), 0)

    def _calc_signal(self, s, a, signal):
        return self.alpha * (signal - self.q[s, a])

    def finish(self, reward):
        self._record.record_reward(reward)
        while not self._record.is_empty:
            g = self._calc_recorded_rewards_sum()
            s, a = next(self._record)
            self.q.learn(s, a, self._calc_signal(s, a, g))


import random

from tensorflow.python.keras._impl.keras.layers import Dense
from tensorflow.python.keras._impl.keras.models import Sequential

from .q_model import QModel


class QRegressionKerasModel(QModel):
    def __init__(self, input_size, hidden_layers):
        QModel.__init__(self, input_size)
        self.input_size = input_size
        self.regressor = Sequential()
        self.regressor.add(Dense(hidden_layers[0], input_dim=input_size, activation="relu"))
        for i in range(1, len(hidden_layers)):
            self.regressor.add(Dense(hidden_layers[i], activation="relu"))
        self.regressor.add(Dense(1, activation="linear"))
        self.regressor.compile("sgd", "mean_squared_error")
        self.is_untrained = True

    def do_prediction(self, state):
        if self.is_untrained:
            return [2 * random.random() - 1]
        predictions = self.regressor.predict(state)
        return predictions

    def do_training(self, states, targets):
        self.regressor.fit(states, targets, batch_size=len(states), epochs=1)
        self.is_untrained = False



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from codecs import open
from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read().replace('\r\n', '\n')

setup(
    name='reinforcement',
    version='1.0.5',
    description='A reinforcement learning module',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/SwamyDev/reinforcement',
    author='Bernhard Raml',
    author_email='pypi-reinforcment@googlegroups.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    keywords='AI reinforcement learning',
    packages=find_packages(exclude=['build', 'dist', 'reinforcement.egg-info', 'test_integration', 'tests']),
    install_requires=['numpy', 'tensorflow'],
    extras_require={
        'test': ['pytest', 'coverage'],
    },
    project_urls={
        'Bug Reports': 'https://github.com/SwamyDev/reinforcement/issues',
        'Source': 'https://github.com/SwamyDev/reinforcement/',
    },
)